@@ -37,13 +37,18 @@ struct DGifCloser {
 };
 using GifUniquePtr = std::unique_ptr<GifFileType, DGifCloser>;
 
+struct PackedRgba {
+  uint8_t r, g, b, a;
+};
+
 // Gif does not support partial transparency, so this considers anything non-0
 // as opaque.
-bool AllOpaque(const ImageF& alpha) {
-  for (size_t y = 0; y < alpha.ysize(); ++y) {
-    const float* const JXL_RESTRICT row = alpha.ConstRow(y);
-    for (size_t x = 0; x < alpha.xsize(); ++x) {
-      if (row[x] == 0.f) {
+bool AllOpaque(const PackedImage& color) {
+  for (size_t y = 0; y < color.ysize; ++y) {
+    const PackedRgba* const JXL_RESTRICT row =
+        static_cast<const PackedRgba*>(color.pixels()) + y * color.xsize;
+    for (size_t x = 0; x < color.xsize; ++x) {
+      if (row[x].a == 0) {
         return false;
       }
     }
@@ -54,7 +59,8 @@ bool AllOpaque(const ImageF& alpha) {
 }  // namespace
 
 Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
-                      ThreadPool* pool, CodecInOut* io) {
+                      const SizeConstraints& constraints,
+                      PackedPixelFile* ppf) {
   int error = GIF_OK;
   ReadState state = {bytes};
   const auto ReadFromSpan = [](GifFileType* const gif, GifByteType* const bytes,
@@ -92,23 +98,21 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
   msan::UnpoisonMemory(gif->SavedImages,
                        sizeof(*gif->SavedImages) * gif->ImageCount);
 
-  const SizeConstraints* constraints = &io->constraints;
-
   JXL_RETURN_IF_ERROR(
-      VerifyDimensions<uint32_t>(constraints, gif->SWidth, gif->SHeight));
+      VerifyDimensions<uint32_t>(&constraints, gif->SWidth, gif->SHeight));
   uint64_t total_pixel_count =
       static_cast<uint64_t>(gif->SWidth) * gif->SHeight;
   for (int i = 0; i < gif->ImageCount; ++i) {
     const SavedImage& image = gif->SavedImages[i];
     uint32_t w = image.ImageDesc.Width;
     uint32_t h = image.ImageDesc.Height;
-    JXL_RETURN_IF_ERROR(VerifyDimensions<uint32_t>(constraints, w, h));
+    JXL_RETURN_IF_ERROR(VerifyDimensions<uint32_t>(&constraints, w, h));
     uint64_t pixel_count = static_cast<uint64_t>(w) * h;
     if (total_pixel_count + pixel_count < total_pixel_count) {
       return JXL_FAILURE("Image too big");
     }
     total_pixel_count += pixel_count;
-    if (total_pixel_count > constraints->dec_max_pixels) {
+    if (total_pixel_count > constraints.dec_max_pixels) {
       return JXL_FAILURE("Image too big");
     }
   }
@@ -122,23 +126,34 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
   }
 
   if (gif->ImageCount > 1) {
-    io->metadata.m.have_animation = true;
+    ppf->info.have_animation = true;
     // Delays in GIF are specified in 100ths of a second.
-    io->metadata.m.animation.tps_numerator = 100;
+    ppf->info.animation.tps_numerator = 100;
+    ppf->info.animation.tps_denominator = 1;
   }
 
-  io->frames.clear();
-  io->frames.reserve(gif->ImageCount);
-  io->dec_pixels = 0;
+  ppf->frames.clear();
+  ppf->frames.reserve(gif->ImageCount);
 
-  io->metadata.m.SetUintSamples(8);
-  io->metadata.m.SetAlphaBits(0);
+  ppf->info.xsize = gif->SWidth;
+  ppf->info.ysize = gif->SHeight;
+  ppf->info.bits_per_sample = 8;
+  ppf->info.exponent_bits_per_sample = 0;
+  // alpha_bits is later set to 8 if we find a frame with transparent pixels.
+  ppf->info.alpha_bits = 0;
+  ppf->info.alpha_exponent_bits = 0;
   JXL_RETURN_IF_ERROR(ApplyColorHints(color_hints, /*color_already_set=*/false,
-                                      /*is_gray=*/false, io));
+                                      /*is_gray=*/false, ppf));
+
+  ppf->info.num_color_channels = 3;
+
+  const JxlPixelFormat format{
+      /*num_channels=*/4u,
+      /*data_type=*/JXL_TYPE_UINT8,
+      /*endianness=*/JXL_NATIVE_ENDIAN,
+      /*align=*/0,
+  };
 
-  Image3F canvas(gif->SWidth, gif->SHeight);
-  io->SetSize(gif->SWidth, gif->SHeight);
-  ImageF alpha(gif->SWidth, gif->SHeight);
   GifColorType background_color;
   if (gif->SColorMap == nullptr ||
       gif->SBackGroundColor >= gif->SColorMap->ColorCount) {
@@ -146,10 +161,12 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
   } else {
     background_color = gif->SColorMap->Colors[gif->SBackGroundColor];
   }
-  FillPlane<float>(background_color.Red, &canvas.Plane(0));
-  FillPlane<float>(background_color.Green, &canvas.Plane(1));
-  FillPlane<float>(background_color.Blue, &canvas.Plane(2));
-  ZeroFillImage(&alpha);
+  const PackedRgba background_rgba{background_color.Red, background_color.Green,
+                                   background_color.Blue, 0};
+  PackedFrame canvas(gif->SWidth, gif->SHeight, format);
+  std::fill_n(static_cast<PackedRgba*>(canvas.color.pixels()),
+              canvas.color.xsize * canvas.color.ysize, background_rgba);
+  Rect canvas_rect{0, 0, canvas.color.xsize, canvas.color.ysize};
 
   Rect previous_rect_if_restore_to_background;
 
@@ -163,7 +180,7 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
                                                image.ImageDesc.Height);
     const Rect image_rect(image.ImageDesc.Left, image.ImageDesc.Top,
                           image.ImageDesc.Width, image.ImageDesc.Height);
-    io->dec_pixels += image_rect.xsize() * image_rect.ysize();
+
     Rect total_rect;
     if (previous_rect_if_restore_to_background.xsize() != 0 ||
         previous_rect_if_restore_to_background.ysize() != 0) {
@@ -186,9 +203,14 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
       total_rect = image_rect;
       replace = false;
     }
-    if (!image_rect.IsInside(canvas)) {
+    if (!image_rect.IsInside(canvas_rect)) {
       return JXL_FAILURE("GIF frame extends outside of the canvas");
     }
+
+    // Allocates the frame buffer.
+    ppf->frames.emplace_back(total_rect.xsize(), total_rect.ysize(), format);
+    auto* frame = &ppf->frames.back();
+
     const ColorMapObject* const color_map =
         image.ImageDesc.ColorMap ? image.ImageDesc.ColorMap : gif->SColorMap;
     JXL_CHECK(color_map);
@@ -199,20 +221,19 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
     DGifSavedExtensionToGCB(gif.get(), i, &gcb);
     msan::UnpoisonMemory(&gcb, sizeof(gcb));
 
-    ImageBundle bundle(&io->metadata.m);
-    if (io->metadata.m.have_animation) {
-      bundle.duration = gcb.DelayTime;
-      bundle.origin.x0 = total_rect.x0();
-      bundle.origin.y0 = total_rect.y0();
+    if (ppf->info.have_animation) {
+      frame->frame_info.duration = gcb.DelayTime;
+      frame->x0 = image_rect.x0();
+      frame->y0 = image_rect.y0();
       if (last_base_was_none) {
         replace = true;
       }
-      bundle.blend = !replace;
+      frame->blend = !replace;
       // TODO(veluca): this could in principle be implemented.
       if (last_base_was_none &&
           (total_rect.x0() != 0 || total_rect.y0() != 0 ||
-           total_rect.xsize() != canvas.xsize() ||
-           total_rect.ysize() != canvas.ysize() || !replace)) {
+           total_rect.xsize() != canvas.color.xsize ||
+           total_rect.ysize() != canvas.color.ysize || !replace)) {
         return JXL_FAILURE(
             "GIF with dispose-to-0 is not supported for non-full or "
             "blended frames");
@@ -220,101 +241,97 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
       switch (gcb.DisposalMode) {
         case DISPOSE_DO_NOT:
         case DISPOSE_BACKGROUND:
-          bundle.use_for_next_frame = true;
+          frame->use_for_next_frame = true;
           last_base_was_none = false;
           break;
         case DISPOSE_PREVIOUS:
-          bundle.use_for_next_frame = false;
+          frame->use_for_next_frame = false;
           break;
         default:
-          bundle.use_for_next_frame = false;
+          frame->use_for_next_frame = false;
           last_base_was_none = true;
       }
     }
-    Image3F frame = CopyImage(canvas);
-    ImageF frame_alpha = CopyImage(alpha);
+
+    // Update the canvas by creating a copy first.
+    PackedImage new_canvas_image(canvas.color.xsize, canvas.color.ysize,
+                                 canvas.color.format);
+    memcpy(new_canvas_image.pixels(), canvas.color.pixels(),
+           new_canvas_image.pixels_size);
     for (size_t y = 0, byte_index = 0; y < image_rect.ysize(); ++y) {
-      float* const JXL_RESTRICT row_r = image_rect.Row(&frame.Plane(0), y);
-      float* const JXL_RESTRICT row_g = image_rect.Row(&frame.Plane(1), y);
-      float* const JXL_RESTRICT row_b = image_rect.Row(&frame.Plane(2), y);
-      float* const JXL_RESTRICT row_alpha = image_rect.Row(&frame_alpha, y);
+      // Assumes format.align == 0. row points to the beginning of the y row in
+      // the image_rect.
+      PackedRgba* row = static_cast<PackedRgba*>(new_canvas_image.pixels()) +
+                        (y + image_rect.y0()) * new_canvas_image.xsize +
+                        image_rect.x0();
       for (size_t x = 0; x < image_rect.xsize(); ++x, ++byte_index) {
         const GifByteType byte = image.RasterBits[byte_index];
         if (byte >= color_map->ColorCount) {
           return JXL_FAILURE("GIF color is out of bounds");
         }
+
         if (byte == gcb.TransparentColor) continue;
         GifColorType color = color_map->Colors[byte];
-        row_alpha[x] = 1.f;
-        row_r[x] = (1.f / 255) * color.Red;
-        row_g[x] = (1.f / 255) * color.Green;
-        row_b[x] = (1.f / 255) * color.Blue;
+        row[x].r = color.Red;
+        row[x].g = color.Green;
+        row[x].b = color.Blue;
+        row[x].a = 255;
       }
     }
-    Image3F sub_frame(total_rect.xsize(), total_rect.ysize());
-    ImageF sub_frame_alpha(total_rect.xsize(), total_rect.ysize());
+    const PackedImage& sub_frame_image = frame->color;
     bool blend_alpha = false;
     if (replace) {
-      CopyImageTo(total_rect, frame, &sub_frame);
-      CopyImageTo(total_rect, frame_alpha, &sub_frame_alpha);
+      // Copy from the new canvas image to the subframe
+      for (size_t y = 0; y < total_rect.ysize(); ++y) {
+        const PackedRgba* row_in =
+            static_cast<const PackedRgba*>(new_canvas_image.pixels()) +
+            (y + total_rect.y0()) * new_canvas_image.xsize + total_rect.x0();
+        PackedRgba* row_out =
+            static_cast<PackedRgba*>(sub_frame_image.pixels()) +
+            y * sub_frame_image.xsize;
+        memcpy(row_out, row_in, sub_frame_image.xsize * sizeof(PackedRgba));
+      }
     } else {
       for (size_t y = 0, byte_index = 0; y < image_rect.ysize(); ++y) {
-        float* const JXL_RESTRICT row_r = sub_frame.PlaneRow(0, y);
-        float* const JXL_RESTRICT row_g = sub_frame.PlaneRow(1, y);
-        float* const JXL_RESTRICT row_b = sub_frame.PlaneRow(2, y);
-        float* const JXL_RESTRICT row_alpha = sub_frame_alpha.Row(y);
+        // Assumes format.align == 0
+        PackedRgba* row = static_cast<PackedRgba*>(sub_frame_image.pixels()) +
+                          y * sub_frame_image.xsize;
         for (size_t x = 0; x < image_rect.xsize(); ++x, ++byte_index) {
           const GifByteType byte = image.RasterBits[byte_index];
           if (byte > color_map->ColorCount) {
             return JXL_FAILURE("GIF color is out of bounds");
           }
           if (byte == gcb.TransparentColor) {
-            row_alpha[x] = 0;
-            row_r[x] = 0;
-            row_g[x] = 0;
-            row_b[x] = 0;
+            row[x].r = 0;
+            row[x].g = 0;
+            row[x].b = 0;
+            row[x].a = 0;
             blend_alpha =
                 true;  // need to use alpha channel if BlendMode blend is used
             continue;
           }
           GifColorType color = color_map->Colors[byte];
-          row_alpha[x] = 1.f;
-          row_r[x] = (1.f / 255) * color.Red;
-          row_g[x] = (1.f / 255) * color.Green;
-          row_b[x] = (1.f / 255) * color.Blue;
+          row[x].r = color.Red;
+          row[x].g = color.Green;
+          row[x].b = color.Blue;
+          row[x].a = 255;
         }
       }
     }
-    bundle.SetFromImage(std::move(sub_frame), ColorEncoding::SRGB());
-    if (has_alpha || !AllOpaque(frame_alpha) || blend_alpha) {
-      if (!has_alpha) {
-        has_alpha = true;
-        io->metadata.m.SetAlphaBits(8);
-        for (ImageBundle& previous_frame : io->frames) {
-          ImageF previous_alpha(previous_frame.xsize(), previous_frame.ysize());
-          FillImage(1.f, &previous_alpha);
-          previous_frame.SetAlpha(std::move(previous_alpha),
-                                  /*alpha_is_premultiplied=*/false);
-        }
-      }
-      bundle.SetAlpha(std::move(sub_frame_alpha),
-                      /*alpha_is_premultiplied=*/false);
+
+    if (!has_alpha && (!AllOpaque(sub_frame_image) || blend_alpha)) {
+      has_alpha = true;
+      ppf->info.alpha_bits = 8;
     }
-    io->frames.push_back(std::move(bundle));
+
     switch (gcb.DisposalMode) {
       case DISPOSE_DO_NOT:
-        canvas = std::move(frame);
-        alpha = std::move(frame_alpha);
+        canvas.color = std::move(new_canvas_image);
         break;
 
       case DISPOSE_BACKGROUND:
-        FillPlane<float>((1.f / 255) * background_color.Red, &canvas.Plane(0),
-                         image_rect);
-        FillPlane<float>((1.f / 255) * background_color.Green, &canvas.Plane(1),
-                         image_rect);
-        FillPlane<float>((1.f / 255) * background_color.Blue, &canvas.Plane(2),
-                         image_rect);
-        FillPlane(0.f, &alpha, image_rect);
+        std::fill_n(static_cast<PackedRgba*>(canvas.color.pixels()),
+                    canvas.color.xsize * canvas.color.ysize, background_rgba);
         previous_rect_if_restore_to_background = image_rect;
         break;
 
@@ -323,15 +340,11 @@ Status DecodeImageGIF(Span<const uint8_t> bytes, const ColorHints& color_hints,
 
       case DISPOSAL_UNSPECIFIED:
       default:
-        FillPlane<float>((1.f / 255) * background_color.Red, &canvas.Plane(0));
-        FillPlane<float>((1.f / 255) * background_color.Green,
-                         &canvas.Plane(1));
-        FillPlane<float>((1.f / 255) * background_color.Blue, &canvas.Plane(2));
-        ZeroFillImage(&alpha);
+        std::fill_n(static_cast<PackedRgba*>(canvas.color.pixels()),
+                    canvas.color.xsize * canvas.color.ysize, background_rgba);
     }
   }
 
-  SetIntensityTarget(io);
   return true;
 }
 
