@@ -28,6 +28,7 @@
 #include "lib/jxl/luminance.h"
 
 namespace jxl {
+namespace extras {
 namespace {
 
 uint64_t get_be_int(int bytes, const uint8_t*& pos, const uint8_t* maxpos) {
@@ -187,7 +188,8 @@ Status decode_layer(const uint8_t*& pos, const uint8_t* maxpos,
 
 }  // namespace
 
-Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
+Status DecodeImagePSD(const Span<const uint8_t> bytes,
+                      const ColorHints& /*color_hints*/, ThreadPool* pool,
                       CodecInOut* io) {
   const uint8_t* pos = bytes.data();
   const uint8_t* maxpos = bytes.data() + bytes.size();
@@ -229,6 +231,7 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
   bool hasmergeddata = true;
   bool have_alpha = false;
   bool merged_has_alpha = false;
+  bool color_already_set = false;
   size_t metalength = get_be_int(4, pos, maxpos);
   const uint8_t* metaoffset = pos;
   while (pos < metaoffset + metalength) {
@@ -257,6 +260,7 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
       if (!io->metadata.m.color_encoding.SetICC(std::move(icc))) {
         return JXL_FAILURE("PSD: Invalid color profile");
       }
+      color_already_set = true;
     } else if (id == 1057) {  // compatibility mode or not?
       if (get_be_int(4, pos, maxpos) != 1) {
         return JXL_FAILURE("PSD: expected version=1 in id=1057 resource block");
@@ -310,6 +314,11 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
     if (blocklength & 1) pos++;  // padding again
   }
 
+  // TODO(deymo): Apply color hints when PSD is converted to PackedPixelFile.
+  (void)color_already_set;
+  // JXL_RETURN_IF_ERROR(ApplyColorHints(color_hints, color_already_set,
+  //                                     /*is_gray=*/false, io));
+
   size_t layerlength = get_be_int(4 * version, pos, maxpos);
   const uint8_t* after_layers_pos = pos + layerlength;
   if (after_layers_pos < pos) return JXL_FAILURE("PSD: invalid layer length");
@@ -349,6 +358,7 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
             return JXL_FAILURE("PSD: Invalid layer count");
           }
           JXL_DEBUG_V(PSD_VERBOSITY, "Real layer count: %i", layercount);
+          if (layercount > 1) have_alpha = true;
           break;
         }
         if (!safe_strncmp(tpos, maxpos, "Mtrn", 4) ||
@@ -356,7 +366,6 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
             !safe_strncmp(tpos, maxpos, "Mt32", 4)) {
           JXL_DEBUG_V(PSD_VERBOSITY, "Merged layer has transparency channel");
           if (nb_channels > real_nb_channels) {
-            real_nb_channels++;
             have_alpha = true;
             merged_has_alpha = true;
           }
@@ -366,14 +375,12 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
     } else if (layercount < 0) {
       // negative layer count indicates merged has alpha and it is to be shown
       if (nb_channels > real_nb_channels) {
-        real_nb_channels++;
         have_alpha = true;
         merged_has_alpha = true;
       }
       layercount = -layercount;
     } else {
       // multiple layers implies there is alpha
-      real_nb_channels++;
       have_alpha = true;
     }
 
@@ -387,6 +394,7 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
     }
     if (have_alpha) {
       JXL_DEBUG_V(PSD_VERBOSITY, "Have alpha");
+      real_nb_channels++;
       info.type = ExtraChannel::kAlpha;
       info.alpha_associated =
           false;  // true? PSD is not consistent with this, need to check
@@ -575,7 +583,7 @@ Status DecodeImagePSD(const Span<const uint8_t> bytes, ThreadPool* pool,
       return JXL_FAILURE("Inconsistent layer configuration");
     }
     if (!merged_has_alpha) {
-      if (colormodel <= real_nb_channels) {
+      if (colormodel >= real_nb_channels) {
         return JXL_FAILURE("Inconsistent layer configuration");
       }
       chan_id.erase(chan_id.begin() + colormodel);
@@ -607,4 +615,5 @@ Status EncodeImagePSD(const CodecInOut* io, const ColorEncoding& c_desired,
   return JXL_FAILURE("PSD encoding not yet implemented");
 }
 
+}  // namespace extras
 }  // namespace jxl
